from performanceplatform.collector.ga.plugins.department import \
    try_get_department

from performanceplatform.collector.ga import \
    create_client, query_documents_for, send_data

from performanceplatform.collector.write import DataSet


def main(credentials, data_set_config, query, options, start_at, end_at):
    client = create_client(credentials)

    assert "filtersets" in options, "`filtersets` must be specified"

    documents = []

    original_filters = query.get("filters", [])

    def insert_ga(filters):
        # Insert ga: on everything but the first.
        def gen():
            for i, f in enumerate(filters):
                if i == 0:
                    yield f
                else:
                    yield "ga:" + f

        return list(gen())

    for filters in options["filtersets"]:
        all_filters = insert_ga(original_filters + filters)
        query["filters"] = [";".join(all_filters)]

        for f in filters:
            if f.startswith("customVarValue9=~^"):
                department = try_get_department(f[len("customVarValue9=~^"):])
                break
        else:
            # If no matches, raise a RuntimeError
            raise RuntimeError("department not found in filters, expected "
                               "filter expression beginning "
                               "'customVarValue9=~^'.")

        options.setdefault("additionalFields", {}).update(
            {"department": department})

        ds = query_documents_for(
            client, query, options, data_set_config['data-type'],
            start_at, end_at
        )

        documents.extend(ds)

    send_data(DataSet.from_config(data_set_config), documents)



# encoding: utf-8

from __future__ import unicode_literals

import json
import logging
import re

import scraperwiki
from dshelpers import batch_processor

from performanceplatform.collector.write import JsonEncoder

from sales_parser import process_csv
from table_names import _RAW_SALES_TABLE, _AGGREGATED_SALES_TABLE
from aggregate import calculate_aggregated_sales, make_spending_group_keys


def nuke_local_database():
    for table in (_RAW_SALES_TABLE, _AGGREGATED_SALES_TABLE):
        _clear_database_table(table)


def save_raw_data(fobj):
    def save(rows):
        scraperwiki.sqlite.save(
            unique_keys=[],
            data=rows,
            table_name=_RAW_SALES_TABLE)

    with batch_processor(save) as b:
        for row in process_csv(fobj):
            logging.debug(','.join(['{}'.format(v) for v in row.values()]))
            b.push(row)


def aggregate_and_save():
    def save(rows):
        scraperwiki.sqlite.save(
            table_name=_AGGREGATED_SALES_TABLE,
            data=rows,
            unique_keys=['_id'])

    with batch_processor(save) as saver:
        for row in calculate_aggregated_sales(make_spending_group_keys()):
            saver.push(row)


def push_aggregates(data_set):
    table = _AGGREGATED_SALES_TABLE

    with batch_processor(data_set.post) as uploader:
        for row in scraperwiki.sqlite.select('* FROM {}'.format(table)):
            uploader.push(row)


def save_to_json():
    table = _AGGREGATED_SALES_TABLE
    with open('{}.json'.format(table), 'w') as f:
        records = [row for row in scraperwiki.sqlite.select(
                   '* FROM {}'.format(table))]

        f.write(json.dumps(records, cls=JsonEncoder, indent=1))


def _clear_database_table(table_name):
    logging.info("Dropping table '{}'".format(table_name))
    scraperwiki.sqlite.execute('DROP TABLE IF EXISTS "{}"'.format(table_name))
    scraperwiki.sqlite.commit()


def normalise(name):
    """
    >>> normalise('Foo  Bar Ltd.')
    u'foo bar ltd'
    >>> normalise('BAZ JIM LTD')
    u'baz jim ltd'
    >>> normalise('BAZ-JIM (South) LTD.')
    u'baz jim south ltd'
    """
    return re.sub('[^a-z]+', ' ', name.lower()).strip()



import datetime
import logging
import requests
import pytz
from performanceplatform.utils import requests_with_backoff
import json


class JsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            if obj.tzinfo is None:
                obj = obj.replace(tzinfo=pytz.UTC)
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


class DataSet(object):
    """Client for writing to a Performance Platform data-set"""

    @staticmethod
    def from_config(config):
        return DataSet(
            config['url'],
            config['token'],
            config['dry_run']
        )

    def __init__(self, url, token, dry_run=False):
        self.url = url
        self.token = token
        self.dry_run = dry_run

    @staticmethod
    def _make_headers(token):
        return {
            "Authorization": "Bearer {}".format(token),
            "Content-type": "application/json"
        }

    @staticmethod
    def _encode_json(data):
        return json.dumps(data, cls=JsonEncoder)

    @staticmethod
    def _log_request(method, url, headers, body):
        logging.info("HTTP {} to '{}'\nheaders: {}\nbody: {}".format(
            method, url, headers, body))

    def post(self, records):
        headers = DataSet._make_headers(self.token)
        json_body = DataSet._encode_json(records)

        if self.dry_run:
            DataSet._log_request('POST', self.url, headers, json_body)
        else:
            if False and len(json_body) > 2048:
                # compress the request
                headers["Content-Encoding"] = "gzip"
                import gzip
                from io import BytesIO
                bio = BytesIO()
                f = gzip.GzipFile(filename='', mode='wb', fileobj=bio)
                f.write(json_body)
                f.close()
                json_body = bio

            response = requests_with_backoff.post(
                url=self.url,
                headers=headers,
                data=json_body
            )

            try:
                response.raise_for_status()
            except:
                logging.error('[PP: {}]\n{}'.format(
                    self.url,
                    response.text))
                raise

            logging.debug("[PP] " + response.text)

    def empty_data_set(self):
        headers = DataSet._make_headers(self.token)
        json_body = DataSet._encode_json([])

        if self.dry_run:
            DataSet._log_request('PUT', self.url, headers, json_body)
        else:
            response = requests.put(
                url=self.url, headers=headers, data=json_body)
            response.raise_for_status()



import os
import logging
import importlib

from collections import OrderedDict

from performanceplatform.collector import arguments
from performanceplatform.collector.logging_setup import set_up_logging


def logging_for_entrypoint(entrypoint, json_fields):
    logfile_path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), '..', '..', 'log')
    set_up_logging(entrypoint, logging.INFO, logfile_path, json_fields)


def merge_performanceplatform_config(
        performanceplatform, data_set, token, dry_run=False):
    return {
        'url': '{0}/{1}/{2}'.format(
            performanceplatform['url'],
            data_set['data-group'],
            data_set['data-type']
        ),
        'token': token['token'],
        'data-group': data_set['data-group'],
        'data-type': data_set['data-type'],
        'dry_run': dry_run
    }


def make_extra_json_fields(args):
    """
    From the parsed command-line arguments, generate a dictionary of additional
    fields to be inserted into JSON logs (logstash_formatter module)
    """
    return {
        'data_group': _get_data_group(args.query),
        'data_type': _get_data_type(args.query),
        'data_group_data_type': _get_data_group_data_type(args.query),
        'query': _get_query_params(args.query),
    }


def _get_data_group(query):
    return query['data-set']['data-group']


def _get_data_type(query):
    return query['data-set']['data-type']


def _get_data_group_data_type(query):
    return '{}/{}'.format(_get_data_group(query), _get_data_type(query))


def _get_query_params(query):
    """
    >>> _get_query_params({'query': {'a': 1, 'c': 3, 'b': 5}})
    'a=1 b=5 c=3'
    """
    query_params = OrderedDict(sorted(query['query'].items()))
    return ' '.join(['{}={}'.format(k, v) for k, v in query_params.items()])


def main():
    args = arguments.parse_args('Performance Platform Collector')

    entrypoint = args.query['entrypoint']

    if args.console_logging:
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)
    else:
        logging_for_entrypoint(entrypoint, make_extra_json_fields(args))

    entrypoint_module = importlib.import_module(entrypoint)
    entrypoint_module.main(
        args.credentials,
        merge_performanceplatform_config(
            args.performanceplatform,
            args.query['data-set'],
            args.token,
            args.dry_run
        ),
        args.query['query'],
        args.query['options'],
        args.start_at,
        args.end_at
    )


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# encoding: utf-8

from __future__ import unicode_literals

import itertools
import datetime
import logging

from collections import OrderedDict, namedtuple

import scraperwiki

from table_names import _RAW_SALES_TABLE

_SECTOR_NAME = {
    'central-gov':         'Central government',
    'local-gov':           'Local government',
    'not-for-profit':      'Not for profit',
    'wider-public-sector': 'Other wider public sector',
    'unknown-sector':      'Unknown',
}

_SME_LARGE_NAME = {
    'sme':               'Small and medium enterprises',
    'large':             'Large enterprises',
}

_LOT_NAME = {
    'iaas': 'Cloud Infrastructure as a Service (IaaS)',
    'paas': 'Cloud Platform as a Service (PaaS)',
    'saas': 'Cloud Software as a Service (SaaS)',
    'css':  'Cloud Support Services (CSS)',
}


class SpendingGroupKey(namedtuple('SpendingGroupKey',
                                  'month,year,lot,sector,sme_large')):
    """
    A 'spending group' is a specific combination of month, year, lot (type),
    government sector and sme/large.
    """

    def __str__(self):
        return str(self.key())

    def __unicode__(self):
        return unicode(self.key())

    def key(self):
        return "{year}_{month:02d}_lot{lot}_{sector}_{sme_large}".format(
            year=self.year, month=self.month, lot=self.lot, sector=self.sector,
            sme_large=self.sme_large)


class CompanyTypeKey(namedtuple('CompanyTypeKey',
                                'month,year,sme_large')):
    def __str__(self):
        return str(self.key())

    def __unicode__(self):
        return unicode(self.key())

    def key(self):
        return "{year}_{month:02d}_{sme_large}".format(
            year=self.year, month=self.month, sme_large=self.sme_large)


def calculate_aggregated_sales(keys):
    for key in keys:
        monthly_spend, transaction_count = get_monthly_spend_and_count(key)
        cumulative_spend = get_cumulative_spend(key)
        cumulative_count = get_cumulative_count(key)
        yield make_aggregated_sales_row(key, monthly_spend, transaction_count,
                                        cumulative_spend, cumulative_count)


def make_aggregated_sales_row(key, monthly_total, transaction_count,
                              cumulative_spend, cumulative_count):
    return OrderedDict([
        ('_id', unicode(key)),
        ('_timestamp', datetime.datetime(key.year, key.month, 1)),
        ('lot', _LOT_NAME[key.lot]),
        ('sector', _SECTOR_NAME[key.sector]),
        ('sme_large', _SME_LARGE_NAME[key.sme_large]),
        ('monthly_spend', monthly_total),
        ('count', transaction_count),
        ('cumulative_spend', cumulative_spend),
        ('cumulative_count', cumulative_count),
    ])


def get_distinct_month_years():
    for row in scraperwiki.sqlite.select(
            'month, year FROM {table} GROUP BY year, month'
            ' ORDER BY year, month'.format(table=_RAW_SALES_TABLE)):
        yield (row['month'], row['year'])


def get_monthly_spend_and_count(key):
    query = ('ROUND(SUM(total_ex_vat), 2) AS total_spend, '
             'COUNT(*) AS invoice_count '
             'FROM {table} '
             'WHERE year={year} '
             'AND month={month} '
             'AND lot="{lot}" '
             'AND customer_sector="{sector}" '
             'AND supplier_type="{sme_large}"'.format(
                 table=_RAW_SALES_TABLE,
                 year=key.year,
                 month=key.month,
                 lot=key.lot,
                 sector=key.sector,
                 sme_large=key.sme_large))
    logging.debug(query)
    result = scraperwiki.sqlite.select(query)[0]
    logging.debug(result)

    spend, count = 0.0, 0

    if result['total_spend'] is not None:
        spend = float(result['total_spend'])

    if result['invoice_count'] is not None:
        count = int(result['invoice_count'])

    return (spend, count)


def get_cumulative_spend(key):
    """
    Get the sum of spending for this category up to and including the given
    month.
    """
    query = ('ROUND(SUM(total_ex_vat), 2) AS total '
             'FROM {table} '
             'WHERE date <= "{year}-{month:02}-01" '
             'AND lot="{lot}" '
             'AND customer_sector="{sector}" '
             'AND supplier_type="{sme_large}"'.format(
                 table=_RAW_SALES_TABLE,
                 year=key.year,
                 month=key.month,
                 lot=key.lot,
                 sector=key.sector,
                 sme_large=key.sme_large))
    logging.debug(query)
    result = scraperwiki.sqlite.select(query)
    logging.debug(result)
    value = result[0]['total']
    return float(result[0]['total']) if value is not None else 0.0


def get_cumulative_count(key):
    """
    Get the sum of spending for this category up to and including the given
    month.
    """
    query = ('COUNT(*) AS total '
             'FROM {table} '
             'WHERE date <= "{year}-{month:02}-01" '
             'AND lot="{lot}" '
             'AND customer_sector="{sector}" '
             'AND supplier_type="{sme_large}"'.format(
                 table=_RAW_SALES_TABLE,
                 year=key.year,
                 month=key.month,
                 lot=key.lot,
                 sector=key.sector,
                 sme_large=key.sme_large))
    logging.debug(query)
    result = scraperwiki.sqlite.select(query)
    logging.debug(result)
    value = result[0]['total']
    return float(result[0]['total']) if value is not None else 0


def make_spending_group_keys():

    month_and_years = get_distinct_month_years()

    all_lots = _LOT_NAME.keys()              # ie [1, 2, 3, 4]
    all_sectors = _SECTOR_NAME.keys()        # ie ['central-gov', 'local-gov']
    all_sme_large = _SME_LARGE_NAME.keys()   # ie ['sme', 'large']

    for (month, year), lot, sector, sme_large in itertools.product(
            month_and_years, all_lots, all_sectors, all_sme_large):

        yield SpendingGroupKey(month=month, year=year, lot=lot, sector=sector,
                               sme_large=sme_large)


def make_company_type_keys():
    month_and_years = get_distinct_month_years()

    all_sme_large = _SME_LARGE_NAME.keys()   # ie ['sme', 'large']
    for (month, year), sme_large in itertools.product(
            month_and_years, all_sme_large):
        yield CompanyTypeKey(month=month, year=year, sme_large=sme_large)



from httplib2 import *
from httplib2 import DEFAULT_MAX_REDIRECTS
import logging
import time

_MAX_RETRIES = 5


class HttpWithBackoff(Http):

    def request(self,
                uri,
                method="GET",
                body=None,
                headers=None,
                redirections=DEFAULT_MAX_REDIRECTS,
                connection_type=None):

        delay = 10

        for n in range(_MAX_RETRIES):
            response = super(HttpWithBackoff, self).request(
                uri,
                method,
                body,
                headers,
                redirections,
                connection_type)
            response_headers = response[0]
            code = response_headers.status
            if code in [403, 502, 503]:
                logging.info('{} request for {} failed with code {}. '
                             'Retrying in {} seconds...'.format(
                                 method, uri, code, delay))
                time.sleep(delay)
                delay *= 2
            else:
                # bad status codes should be handled by the Google API
                # client in the normal way
                return response

        # we made _MAX_RETRIES requests but none worked
        logging.error('Max retries exceeded for {}'.format(uri))
        return response




# encoding: utf-8

from __future__ import unicode_literals

import mock
import unittest

from nose.tools import assert_equal, assert_is_instance
from os.path import join, dirname, abspath

from sales_parser import (process_csv, parse_latest_csv_url,
                          UNKNOWN_SECTOR_KEY)

from aggregate import (
    SpendingGroupKey,
    make_spending_group_keys, make_company_type_keys,
    make_aggregated_sales_row, get_monthly_spend_and_count,
    get_cumulative_spend, get_cumulative_count, get_distinct_month_years)


SAMPLE_DIR = join(dirname(abspath(__file__)), 'sample_data')


class ExtractLatestUrlTestCase(unittest.TestCase):

    def test_parse_latest_csv_url_index_page(self):
        with open(join(SAMPLE_DIR, 'index_page.html'), 'r') as f:
            assert_equal(
                'http://gcloud.civilservice.gov.uk/files/2013/11/'
                'G-Cloud-Total-Spend-22-11-13.csv',
                parse_latest_csv_url(f))


class SpendingDataTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        with open(join(
                SAMPLE_DIR,
                'G-Cloud Spend 10-04-14 v1 For Publication.csv'), 'r') as f:
            cls.rows = list(process_csv(f))

    def test_correct_number_of_rows(self):
        num_rows = len(self.rows)
        assert_equal(9227, num_rows)

    def test_correct_keys(self):
        first_row = self.rows[0]
        assert_equal(
            set([
                'customer_name',
                'customer_sector',
                'supplier_name',
                'supplier_type',
                'lot',
                'month',
                'year',
                'total_ex_vat',
                'date',
            ]),
            set(first_row.keys()))

    def test_correct_number_of_customer_names(self):
        names = set(row['customer_name'] for row in self.rows)
        assert_equal(437, len(names))

    def test_all_valid_customer_sector(self):
        sectors = set(row['customer_sector'] for row in self.rows)
        assert_equal(
            set([
                'central-gov',
                'local-gov',
                'not-for-profit',
                'wider-public-sector',
                'unknown-sector',
            ]),
            sectors)

    def test_expected_number_of_rows_with_unknown_sector(self):
        unknown_sectors = [row['customer_name'] for row in self.rows
                           if row['customer_sector'] == UNKNOWN_SECTOR_KEY]
        assert_equal(7, len(unknown_sectors))

    def test_correct_number_of_supplier_names(self):
        names = set(row['supplier_name'] for row in self.rows)
        assert_equal(300, len(names))

    def test_all_valid_supplier_types(self):
        supplier_types = set(row['supplier_type'] for row in self.rows)
        assert_equal(
            set([
                'sme',
                'large',
            ]),
            supplier_types)

    def test_all_valid_lot_names(self):
        lot_names = set(row['lot'] for row in self.rows)
        assert_equal(set(['iaas', 'paas', 'saas', 'css']), lot_names)

    def test_all_valid_months(self):
        months = set(row['month'] for row in self.rows)
        assert_equal(set([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]), months)

    def test_all_valid_years(self):
        years = set(row['year'] for row in self.rows)
        assert_equal(set([2012, 2013, 2014]), years)

    def test_totals_are_all_floats(self):
        totals = set(row['total_ex_vat'] for row in self.rows)
        for total in totals:
            assert_is_instance(total, float)  # TODO: Decimal?

    def test_totals_add_up_to_correct_amount(self):
        totals = (row['total_ex_vat'] for row in self.rows)
        assert_equal(154635951.7974, sum(totals))


class AggregationTest(unittest.TestCase):
    def test_make_spending_group_keys(self):
        # 5 sectors x 2 sme|large x 4 lots x 6 month+year combos
        with mock.patch('scraperwiki.sqlite.select') as mocked:
            # mustn't make keys for months 3..12 for 2010
            fake_months_years = [
                {'year': 2009, 'month': 9},
                {'year': 2009, 'month': 10},
                {'year': 2009, 'month': 11},
                {'year': 2009, 'month': 12},
                {'year': 2010, 'month': 1},
                {'year': 2010, 'month': 2},
            ]
            mocked.return_value = fake_months_years

            all_keys = list(make_spending_group_keys())
            assert_equal(sorted(list(set(all_keys))), sorted(all_keys))
            assert_equal(
                5 * 2 * 4 * 6,
                len(list(all_keys))
            )

    def test_make_company_type(self):
        # 2 sme|large x 6 month+year combos
        with mock.patch('scraperwiki.sqlite.select') as mocked:
            # mustn't make keys for months 3..12 for 2010
            fake_months_years = [
                {'year': 2009, 'month': 9},
                {'year': 2009, 'month': 10},
                {'year': 2009, 'month': 11},
                {'year': 2009, 'month': 12},
                {'year': 2010, 'month': 1},
                {'year': 2010, 'month': 2},
            ]
            mocked.return_value = fake_months_years

            all_keys = list(make_company_type_keys())
            assert_equal(sorted(list(set(all_keys))), sorted(all_keys))
            assert_equal(
                2 * 6,
                len(list(all_keys))
            )

    def test_distinct_month_years(self):
        with mock.patch('scraperwiki.sqlite.select') as mocked:
            # mustn't make keys for months 3..12 for 2010
            fake_months_years = [
                {'year': 2009, 'month': 9},
                {'year': 2009, 'month': 10},
                {'year': 2009, 'month': 11},
                {'year': 2009, 'month': 12},
                {'year': 2010, 'month': 1},
                {'year': 2010, 'month': 2},
            ]
            mocked.return_value = fake_months_years
            assert_equal(
                [
                    (9, 2009),
                    (10, 2009),
                    (11, 2009),
                    (12, 2009),
                    (1, 2010),
                    (2, 2010),
                ],
                list(get_distinct_month_years()))

    def test_make_aggregated_sales_row(self):
        fake_key = SpendingGroupKey(
            month=1,
            year=2009,
            lot='paas',
            sector='central-gov',
            sme_large='sme')

        result = make_aggregated_sales_row(fake_key, 1234.56, 5, 4567.89, 10)
        assert_equal([
            '_id',
            '_timestamp',
            'lot',
            'sector',
            'sme_large',
            'monthly_spend',
            'count',
            'cumulative_spend',
            'cumulative_count'],
            result.keys())
        # TODO: test the values

    def test_get_monthly_spend_and_count(self):
        fake_key = SpendingGroupKey(
            month=1,
            year=2009,
            lot='paas',
            sector='central-gov',
            sme_large='sme')

        with mock.patch('scraperwiki.sqlite.select') as mocked:
            mocked.return_value = [
                {'total_spend': 1234.56,
                 'invoice_count': 3}]

            spend, count = get_monthly_spend_and_count(fake_key)
            expected_sql = ('ROUND(SUM(total_ex_vat), 2) AS total_spend, '
                            'COUNT(*) AS invoice_count '
                            'FROM raw_sales '
                            'WHERE year=2009 AND month=1 AND lot="paas" '
                            'AND customer_sector="central-gov" '
                            'AND supplier_type="sme"')
            assert_equal(
                [mock.call(expected_sql)],
                mocked.call_args_list)

        assert_equal(1234.56, spend)
        assert_equal(3, count)

    def test_get_cumulative_spend(self):
        fake_key = SpendingGroupKey(
            month=1,
            year=2009,
            lot="paas",
            sector='central-gov',
            sme_large='sme')

        with mock.patch('scraperwiki.sqlite.select') as mocked:
            mocked.return_value = [{'total': 1234.56}]

            spend = get_cumulative_spend(fake_key)
            expected_sql = ('ROUND(SUM(total_ex_vat), 2) AS total '
                            'FROM raw_sales '
                            'WHERE date <= "2009-01-01" '
                            'AND lot="paas" AND customer_sector="central-gov" '
                            'AND supplier_type="sme"')

            assert_equal(
                [mock.call(expected_sql)],
                mocked.call_args_list)

        assert_equal(1234.56, spend)

    def test_get_cumulative_count(self):
        fake_key = SpendingGroupKey(
            month=1,
            year=2009,
            lot='paas',
            sector='central-gov',
            sme_large='sme')

        with mock.patch('scraperwiki.sqlite.select') as mocked:
            mocked.return_value = [{'total': 17}]

            count = get_cumulative_count(fake_key)
            expected_sql = ('COUNT(*) AS total FROM raw_sales '
                            'WHERE date <= "2009-01-01" AND lot="paas" AND '
                            'customer_sector="central-gov" '
                            'AND supplier_type="sme"')

            assert_equal(
                [mock.call(expected_sql)],
                mocked.call_args_list)

        assert_equal(17, count)



from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)

from performanceplatform.collector.ga.core \
    import create_client, query_documents_for, send_data

from performanceplatform.collector.write import DataSet


def main(credentials, data_set_config, query, options, start_at, end_at):
    client = create_client(credentials)

    documents = query_documents_for(
        client, query, options,
        data_set_config['data-type'],
        start_at, end_at)

    data_set = DataSet.from_config(data_set_config)
    send_data(data_set, documents)



import logging
import time

# import everything from requests so that other code can import this package
# instead of requests
from requests import *

_MAX_RETRIES = 5


def get(url, *args, **kwargs):
    return __request_with_backoff('GET', url, *args, **kwargs)


def post(url, *args, **kwargs):
    return __request_with_backoff('POST', url, *args, **kwargs)


def __request_with_backoff(method, url, *args, **kwargs):
    delay = 10

    for n in range(_MAX_RETRIES):
        response = request(method, url, *args, **kwargs)
        code = response.status_code
        if code in [403, 502, 503]:
            logging.info('{} request for {} failed with code {}. Retrying in '
                         '{} seconds...'.format(method, url, code, delay))
            time.sleep(delay)
            delay *= 2
        else:
            response.raise_for_status()
            return response

    # we made _MAX_RETRIES requests but none worked
    logging.error('Max retries exceeded for {}'.format(url))
    response.raise_for_status()



import argparse
import re
import sys


ignore_line_re = re.compile("^#.*|\s*$")


class ParseError(StandardError):
    pass


def crontab_begin_comment(unique_id):
    return '# Begin performanceplatform.collector jobs for %s' % unique_id


def crontab_end_comment(unique_id):
    return '# End performanceplatform.collector jobs for %s' % unique_id


def remove_existing_crontab_for_app(crontab, unique_id):
    new_crontab = []
    should_append = True
    for line in crontab:
        if line == crontab_begin_comment(unique_id):
            should_append = False
        if should_append:
            new_crontab.append(line)
        if line == crontab_end_comment(unique_id):
            should_append = True
    return new_crontab


def parse_job_line(line):
    """
    >>> parse_job_line( \
            "* * * *,myquery,mycredentials,mytoken,performanceplatform\\n")
    ('* * * *', 'myquery', 'mycredentials', 'mytoken', 'performanceplatform')
    >>> parse_job_line("            ") is None
    True
    >>> parse_job_line("# comment") is None
    True
    """
    parsed = None

    if not ignore_line_re.match(line):
        parsed = tuple(line.strip().split(','))

    return parsed


def generate_crontab(current_crontab, path_to_jobs, path_to_app, unique_id):
    """Returns a crontab with jobs from job path

    It replaces jobs previously generated by this function
    It preserves jobs not generated by this function
    """
    job_template = '{schedule} {app_path}/venv/bin/pp-collector ' \
                   '-q {app_path}/config/{query} ' \
                   '-c {app_path}/config/{credentials} ' \
                   '-t {app_path}/config/{token} ' \
                   '-b {app_path}/config/{performanceplatform} ' \
                   '>> {app_path}/log/out.log 2>> {app_path}/log/error.log'

    crontab = [line.strip() for line in current_crontab]
    crontab = remove_existing_crontab_for_app(crontab, unique_id)
    additional_crontab = []
    with open(path_to_jobs) as jobs:
        try:
            for job in jobs:
                parsed = parse_job_line(job)

                if parsed is not None:
                    schedule, query, credentials, \
                        token, performanceplatform = parsed

                    cronjob = job_template.format(
                        schedule=schedule,
                        app_path=path_to_app,
                        query=query,
                        credentials=credentials,
                        token=token,
                        performanceplatform=performanceplatform
                    )

                    additional_crontab.append(cronjob)

        except ValueError as e:
            raise ParseError(str(e))

    if additional_crontab:
        crontab.append(crontab_begin_comment(unique_id))
        crontab.extend(additional_crontab)
        crontab.append(crontab_end_comment(unique_id))

    return crontab


if __name__ == '__main__':
    current_crontab = sys.stdin.readlines()
    try:
        parser = argparse.ArgumentParser()
        parser.add_argument('path_to_app',
                            help='Path to where the application')
        parser.add_argument('path_to_jobs',
                            help='Path to the file where job templates are')
        parser.add_argument('app_unique_id',
                            help='Unique id of the application '
                                 'used to update crontab')

        args = parser.parse_args()

        crontab = generate_crontab(current_crontab,
                                   args.path_to_jobs,
                                   args.path_to_app,
                                   args.app_unique_id)
        sys.stdout.write("\n".join(crontab) + "\n")
        sys.exit(0)
    except StandardError as e:
        sys.stderr.write(str(e))
        sys.stdout.write("\n".join(current_crontab))
        sys.exit(1)



import base64
import json
import logging
import re
from performanceplatform.utils.http_with_backoff import HttpWithBackoff

from gapy.client import from_private_key, from_secrets_file

from performanceplatform.collector.ga.datetimeutil \
    import to_datetime, period_range, to_utc


def create_client(credentials):
    if "CLIENT_SECRETS" in credentials:
        return from_secrets_file(
            credentials['CLIENT_SECRETS'],
            storage_path=credentials['STORAGE_PATH'],
            http_client=HttpWithBackoff(),
        )
    else:
        return from_private_key(
            credentials['ACCOUNT_NAME'],
            private_key_path=credentials['PRIVATE_KEY'],
            storage_path=credentials['STORAGE_PATH'],
            http_client=HttpWithBackoff(),
        )


def query_ga(client, config, start_date, end_date):
    logging.info("Querying GA for data in the period: %s - %s"
                 % (str(start_date), str(end_date)))

    # If maxResults is 0, don't include it in the query.
    # Same for a sort == [].

    maxResults = config.get("maxResults", None)
    if maxResults == 0:
        maxResults = None

    sort = config.get("sort", None)
    if sort == []:
        sort = None

    return client.query.get(
        config["id"].replace("ga:", ""),
        start_date,
        end_date,
        config["metrics"],
        config.get("dimensions"),
        config.get("filters"),
        maxResults,
        sort,
    )


def send_data(data_set, documents):
    if len(documents) == 0:
        logging.info("No data returned with current configuration")
        return

    data_set.post(documents)


def _format(timestamp):
    return to_utc(timestamp).strftime("%Y%m%d%H%M%S")


def value_id(value):
    value_bytes = value.encode('utf-8')
    logging.debug(u"'{0}' ({1})".format(value, type(value)))
    return base64.urlsafe_b64encode(value_bytes), value_bytes


def data_id(data_type, timestamp, period, dimension_values):
    # `dimension_values` may be non-string python types and need to be coerced.
    values = map(unicode, dimension_values)
    slugs = [data_type, _format(timestamp), period] + values
    return value_id("_".join(slugs))


def map_one_to_one_fields(mapping, pairs):
    return dict((mapping.get(key, key), value) for key, value in pairs.items())


def map_multi_value_fields(mapping, pairs):
    multi_value_regexp = '^(.*)_(\d*)$'
    multi_value_delimiter = ':'
    mapped_pairs = {}

    for from_key, to_key in mapping.items():
        multi_value_matches = re.search(multi_value_regexp, from_key)
        if multi_value_matches:
            key = multi_value_matches.group(1)
            index = int(multi_value_matches.group(2))
            multi_value = pairs.get(key)
            if multi_value is None:
                continue

            values = multi_value.split(multi_value_delimiter)
            if index < len(values):
                mapped_pairs[to_key] = values[index]

    return mapped_pairs


def apply_key_mapping(mapping, pairs):
    return dict(map_one_to_one_fields(mapping, pairs).items() +
                map_multi_value_fields(mapping, pairs).items())


def try_number(value):
    """
    Attempt to cast the string `value` to an int, and failing that, a float,
    failing that, raise a ValueError.
    """

    for cast_function in [int, float]:
        try:
            return cast_function(value)
        except ValueError:
            pass

    raise ValueError("Unable to use value as int or float: {0!r}"
                     .format(value))


def build_document(item, data_type,
                   mappings=None, idMapping=None):
    if data_type is None:
        raise ValueError("Must provide a data type")
    if mappings is None:
        mappings = {}
    period = "week"

    if idMapping is not None:
        if isinstance(idMapping, list):
            values_for_id = map(lambda d: item['dimensions'][d], idMapping)
            value_for_id = "".join(values_for_id)
        else:
            value_for_id = item['dimensions'][idMapping]

        (_id, human_id) = value_id(value_for_id)
    else:
        (_id, human_id) = data_id(
            data_type,
            to_datetime(item["start_date"]),
            period,
            item.get('dimensions', {}).values())

    base_properties = {
        "_id": _id,
        "_timestamp": to_datetime(item["start_date"]),
        "humanId": human_id,
        "timeSpan": period,
        "dataType": data_type
    }
    dimensions = apply_key_mapping(
        mappings,
        item.get("dimensions", {})).items()

    metrics = [(key, try_number(value))
               for key, value in item["metrics"].items()]
    return dict(base_properties.items() + dimensions + metrics)


def pretty_print(obj):
    return json.dumps(obj, indent=2)


def build_document_set(results, data_type, mappings, idMapping=None):
    return [build_document(item, data_type, mappings, idMapping)
            for item in results]


def query_for_range(client, query, period_start, period_end):
    items = []
    extend = items.extend

    for start, end in period_range(period_start, period_end):
        extend(query_ga(client, query, start, end))

    return items


def run_plugins(plugins_strings, results):

    last_plugin = plugins_strings[-1]
    if not last_plugin.startswith("ComputeIdFrom"):
        raise RuntimeError("Last plugin must be `ComputeIdFrom` for now. This "
                           "may be changed with further development if "
                           "necessary")

    # Import is here so that it is only required when "plugins" is specified
    from performanceplatform.collector.ga.plugins import load_plugins
    plugins = load_plugins(plugins_strings)

    # Plugins are designed so that their configuration is described in the
    # plugin string. At this point, `plugin` should be a closure which only
    # requires the documents to process.
    for plugin in plugins:
        results = plugin(results)

    return results


def query_documents_for(client, query, options,
                        data_type, start_date, end_date):
    results = query_for_range(client, query, start_date, end_date)

    data_type = options.get('dataType', data_type)
    mappings = options.get("mappings", {})
    idMapping = options.get("idMapping", None)

    docs = build_document_set(results, data_type, mappings, idMapping)

    if "additionalFields" in options:
        additional_fields = options["additionalFields"]
        for doc in docs:
            doc.update(additional_fields)

    if "plugins" in options:
        docs = run_plugins(options["plugins"], docs)

    return docs



import argparse
import json
from dateutil.parser import parse as parse_date


def parse_args(name="", args=None):
    """Parse command line argument for a collector

    Returns an argparse.Namespace with 'config' and 'query' options"""
    def _load_json_file(path):
        with open(path) as f:
            return json.load(f)

    parser = argparse.ArgumentParser(description="%s collector for sending"
                                                 " data to the performance"
                                                 " platform" % name)
    parser.add_argument('-c', '--credentials', dest='credentials',
                        type=_load_json_file,
                        help='JSON file containing credentials '
                             'for the collector',
                        required=True)
    parser.add_argument('-q', '--query', dest='query',
                        type=_load_json_file,
                        help='JSON file containing details '
                             'about the query to make'
                             'against the source API '
                             'and the target data-set',
                        required=True)
    parser.add_argument('-t', '--token', dest='token',
                        type=_load_json_file,
                        help='JSON file containing token '
                             'for the collector',
                        required=True)
    parser.add_argument('-b', '--performanceplatform',
                        dest='performanceplatform',
                        type=_load_json_file,
                        help='JSON file containing the Performance Platform '
                             'config for the collector',
                        required=True)
    parser.add_argument('-s', '--start', dest='start_at',
                        type=parse_date,
                        help='Date to start collection from')
    parser.add_argument('-e', '--end', dest='end_at',
                        type=parse_date,
                        help='Date to end collection')
    parser.add_argument('--console-logging', dest='console_logging',
                        action='store_true',
                        help='Output logging to the console rather than file')
    parser.add_argument('--dry-run', dest='dry_run',
                        action='store_true',
                        help='Instead of pushing to the Performance Platform '
                             'the collector will print out what would have '
                             'been pushed')
    parser.set_defaults(console_logging=False, dry_run=False)
    args = parser.parse_args(args)

    return args



from apiclient.discovery import build
from oauth2client.client import flow_from_clientsecrets
from oauth2client.file import Storage
from oauth2client.tools import run
from datetime import datetime
import pytz

from performanceplatform.collector.write import DataSet
from performanceplatform.utils.http_with_backoff import HttpWithBackoff


GOOGLE_API_SCOPE = "https://www.googleapis.com/auth/analytics"


class Collector(object):
    def __init__(self, credentials):
        self._realtime = Realtime(credentials)

    def send_records_for(self, query, to):
        data_set = DataSet.from_config(to)

        visitor_count = self._realtime.query(query)

        record = self._create_record(visitor_count,
                                     query.get('filters', ''))

        data_set.post(record)

    def _create_record(self, visitor_count, for_url):
        timestamp = _timestamp()
        return {
            "_timestamp": timestamp,
            "_id": timestamp,
            "unique_visitors": visitor_count,
            "for_url": for_url
        }


class Realtime(object):
    def __init__(self, credentials):
        self._authenticate(credentials["CLIENT_SECRETS"],
                           credentials["STORAGE_PATH"])

    def _authenticate(self, client_secrets, storage_path):
        flow = flow_from_clientsecrets(client_secrets, scope=GOOGLE_API_SCOPE)
        storage = Storage(storage_path)
        credentials = storage.get()
        if credentials is None or credentials.invalid:
            credentials = run(flow, storage)

        self.service = build(
            serviceName="analytics",
            version="v3",
            http=credentials.authorize(HttpWithBackoff())
        )

    def execute_ga_query(self, query):
        return self.service.data().realtime().get(
            **query
        ).execute()

    def query(self, query):
        response = self.execute_ga_query(query)

        if "rows" in response:
            visitor_count = int(response["rows"][0][0])
        else:
            visitor_count = 0

        return visitor_count


def _timestamp():
    timezone = pytz.UTC
    timestamp = datetime.now(timezone).replace(microsecond=0)
    return timestamp.isoformat()


def main(credentials, data_set, query, options, start_at, end_at):
    collector = Collector(credentials)
    collector.send_records_for(query, data_set)




class RemoveKey(object):

    """
    Remove all of the specified keys from the input documents.
    """

    def __init__(self, *remove_keys):
        self.remove_keys = remove_keys

    def __call__(self, documents):
        for document in documents:
            for key in self.remove_keys:
                del document[key]
        return documents



# Namespace package: https://docs.python.org/2/library/pkgutil.html
from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)

from .aggregate import AggregateKey, aggregate_count, aggregate_rate
from .comment import Comment
from .compute_id import ComputeIdFrom
from .department import ComputeDepartmentKey, SetDepartment
from .rank import ComputeRank
from .remove_key import RemoveKey

from .load_plugin import load_plugins




class ComputeRank(object):

    def __init__(self, var_name):
        self.var_name = var_name

    def __call__(self, documents):
        for i, document in enumerate(documents, 1):
            document[self.var_name] = i
        return documents



from datetime import datetime
import pytz
from performanceplatform.utils import requests_with_backoff
import time
import logging


def _send_authenticated_pingdom_request(path, user, password, app_key,
                                        url_params):
    response = requests_with_backoff.get(
        url="https://api.pingdom.com/api/2.0/" + path,
        auth=(user, password),
        headers={
            'App-key': app_key
        },
        params=url_params
    )

    response.raise_for_status()

    return response.json()


class Pingdom(object):
    def __init__(self, config):
        self.user = config['user']
        self.password = config['password']
        self.app_key = config['app_key']
        self.API_LOCATION = "https://api.pingdom.com/api/2.0/"

    def _make_request(self, path, url_params={}):
        response = requests_with_backoff.get(
            url=self.API_LOCATION + path,
            auth=(self.user, self.password),
            headers={
                "App-Key": self.app_key
            },
            params=url_params
        )
        return response

    def _build_response(self, response):
        hours = response['summary']['hours']
        new_hours = []
        for hour in hours:
            hour.update({'starttime': datetime.fromtimestamp(
                hour['starttime'],
                tz=pytz.UTC
            )})
            new_hours.append(hour)
        return new_hours

    def stats(self, check_name, start, end):
        app_code = self.check_id(check_name)
        params = {
            "includeuptime": "true",
            "from": time.mktime(start.timetuple()),
            "to": time.mktime(end.timetuple()),
            "resolution": "hour"
        }
        path = "summary.performance/" + str(app_code)

        try:
            return self._build_response(_send_authenticated_pingdom_request(
                path=path,
                user=self.user,
                password=self.password,
                app_key=self.app_key,
                url_params=params
            ))
        except requests_with_backoff.exceptions.HTTPError as e:
            logging.error("Request to pingdom failed: %s" % str(e))

    def check_id(self, name):
        checks = _send_authenticated_pingdom_request(
            path="checks",
            user=self.user,
            password=self.password,
            app_key=self.app_key,
            url_params=None
        )

        check_to_find = [check for check in checks["checks"]
                         if check["name"] == name]

        return check_to_find[0]["id"]


class Collector(object):
    def __init__(self, config):
        pass



"""
load_plugin.py
--------------

Responsible for taking plugin strings and returning plugin callables.

"""

# For the linter
import __builtin__

import performanceplatform.collector.ga.plugins


def load_plugins(plugin_names):
    return [load_plugin(plugin_name) for plugin_name in plugin_names]


def load_plugin(plugin_name):

    expr = compile(plugin_name, "performanceplatform.collector plugin", "eval")

    return eval(expr, __builtin__.__dict__,
                performanceplatform.collector.ga.plugins.__dict__)



import re


class ComputeDepartmentKey(object):

    """
    Adds a 'department' key to a dictionary by looking up the department from
    the specified, key_name. It takes the first department code of form
    <[code]> from document[key_name].
    """

    def __init__(self, key_name):
        self.key_name = key_name

    def __call__(self, documents):
        def compute_department(department):
            assert self.key_name in document, (
                'key "{}" not found "{}"'.format(self.key_name, document))
            department_codes = document[self.key_name]
            department_code = take_first_department_code(department_codes)
            document["department"] = DEPARTMENT_MAPPING.get(
                department_code, department_code)
            return document

        return [compute_department(document) for document in documents]


class SetDepartment(object):

    """
    Adds a 'department' key to a dictionary by using the department specified
    in the constructor. It takes the first department code of form <[code]>,
    and if that doesn't exist it uses the string verbatim.

    This assumes that all of the data being processed belongs to the same
    single department.

    For example:

        SetDepartment("<D3>")
        SetDepartment("Department for fooing the bar")
    """

    def __init__(self, department_or_code):
        self.department_or_code = department_or_code
        self.value = try_get_department(self.department_or_code)

    def __call__(self, documents):
        for document in documents:
            document["department"] = self.value
        return documents


def take_first_department_code(department_codes):
    get_first_re = re.compile("^(<[^>]+>).*$")
    match = get_first_re.match(department_codes)
    assert match is not None
    (department_code, ) = match.groups()
    return department_code


def try_get_department(department_or_code):
    """
    Try to take the first department code, or fall back to string as passed
    """
    try:
        value = take_first_department_code(department_or_code)
    except AssertionError:
        value = department_or_code

    if value in DEPARTMENT_MAPPING:
        value = DEPARTMENT_MAPPING[value]

    return value


DEPARTMENT_MAPPING = {
    "<D1>": "attorney-generals-office",
    "<D2>": "cabinet-office",
    "<D3>": "department-for-business-innovation-skills",
    "<D4>": "department-for-communities-and-local-government",
    "<D5>": "department-for-culture-media-sport",
    "<D6>": "department-for-education",
    "<D7>": "department-for-environment-food-rural-affairs",
    "<D8>": "department-for-international-development",
    "<D9>": "department-for-transport",
    "<D10>": "department-for-work-pensions",
    "<D11>": "department-of-energy-climate-change",
    "<D12>": "department-of-health",
    "<D13>": "foreign-commonwealth-office",
    "<D15>": "hm-treasury",
    "<D16>": "home-office",
    "<D17>": "ministry-of-defence",
    "<D18>": "ministry-of-justice",
    "<D19>": "northern-ireland-office",
    "<D20>": "office-of-the-advocate-general-for-scotland",
    "<D21>": "the-office-of-the-leader-of-the-house-of-commons",
    "<D22>": "office-of-the-leader-of-the-house-of-lords",
    "<D23>": "scotland-office",
    "<D24>": "wales-office",
    "<D25>": "hm-revenue-customs",
    "<D30>": "treasury-solicitor-s-department",
    "<D38>": "ordnance-survey",
    "<D85>": "forestry-commission",
    "<D98>": "the-charity-commission-for-england-and-wales",
    "<D101>": "crown-prosecution-service",
    "<D102>": "food-standards-agency",
    "<D106>": "ofsted",
    "<D108>": "ofgem",
    "<D109>": "office-of-qualifications-and-examinations-regulation",
    "<D110>": "office-of-rail-regulation",
    "<D115>": "serious-fraud-office",
    "<D116>": "uk-statistics-authority",
    "<D117>": "uk-trade-investment",
    "<D240>": "the-water-services-regulation-authority",
    "<D241>": "uk-export-finance",
    "<D303>": "office-for-national-statistics",
    "<D346>": "public-works-loan-board",
    "<D352>": "office-of-fair-trading",
    "<D435>": "supreme-court-of-the-united-kingdom",
    "<EA26>": "companies-house",
    "<EA31>": "uk-space-agency",
    "<EA32>": "insolvency-service",
    "<EA33>": "national-measurement-office",
    "<EA34>": "intellectual-property-office",
    "<EA37>": "fire-service-college",
    "<EA39>": "planning-inspectorate",
    "<EA40>": "queen-elizabeth-ii-conference-centre",
    "<EA41>": "royal-parks",
    "<EA42>": "defence-science-and-technology-laboratory",
    "<EA44>": "defence-support-group",
    "<EA46>": "met-office",
    "<EA47>": "ministry-of-defence-police-and-guarding-agency",
    "<EA49>": "service-children-s-education",
    "<EA50>": "service-personnel-and-veterans-agency",
    "<EA51>": "uk-hydrographic-office",
    "<EA52>": "animal-health-and-veterinary-laboratories-agency",
    "<EA53>": "centre-for-environment-fisheries-and-aquaculture-science",
    "<EA54>": "forest-research",
    "<EA55>": "forest-enterprise-england",
    "<EA56>": "the-food-and-environment-research-agency",
    "<EA58>": "rural-payments-agency",
    "<EA60>": "veterinary-medicines-directorate",
    "<EA61>": "fco-services",
    "<EA62>": "wilton-park",
    "<EA63>": "medicines-and-healthcare-products-regulatory-agency",
    "<EA66>": "identity-and-passport-service",
    "<EA67>": "uk-border-agency",
    "<EA70>": "national-offender-management-service",
    "<EA71>": "the-national-archives",
    "<EA72>": "office-of-the-public-guardian",
    "<EA73>": "hm-courts-and-tribunals-service",
    "<EA74>": "driver-and-vehicle-licensing-agency",
    "<EA75>": "driving-standards-agency",
    "<EA77>": "highways-agency",
    "<EA78>": "maritime-and-coastguard-agency",
    "<EA79>": "vehicle-and-operator-services-agency",
    "<EA80>": "vehicle-certification-agency",
    "<EA82>": "uk-debt-management-office",
    "<EA86>": "skills-funding-agency",
    "<EA87>": "valuation-office-agency",
    "<EA104>": "ns-i",
    "<EA114>": "royal-mint",
    "<EA199>": "environment-agency",
    "<EA242>": "education-funding-agency",
    "<EA243>": "standards-and-testing-agency",
    "<EA245>": "national-college-for-school-leadership",
    "<EA321>": "hm-prison-service",
    "<EA365>": "government-procurement-service",
    "<EA480>": "public-health-england",
    "<EA541>": "national-college-for-teaching-and-leadership",
    "<EA570>": "driver-and-vehicle-standards-agency",
    "<IM320>": "criminal-cases-review-commission",
    "<IM324>": "hm-inspectorate-of-prisons",
    "<IM325>": "hm-inspectorate-of-probation",
    "<IM327>": "prisons-and-probation-ombudsman",
    "<IM329>": "the-legal-ombudsman",
    "<IM332>": "legal-services-board",
    "<IM333>": "judicial-appointments-and-conduct-ombudsman",
    "<IM335>": "independent-commission-for-aid-impact",
    "<IM341>": "house-of-lords-appointments-commission",
    "<OT100>": "the-crown-estate",
    "<OT152>": "industrial-development-advisory-board",
    "<OT204>": "agriculture-and-horticulture-development-board",
    "<OT205>": "sea-fish-industry-authority",
    "<OT214>": "covent-garden-market-authority",
    "<OT216>": "broads-authority",
    "<OT217>": "dartmoor-national-park-authority",
    "<OT219>": "exmoor-national-park-authority",
    "<OT220>": "lake-district-national-park-authority",
    "<OT222>": "new-forest-national-park-authority",
    "<OT223>": "north-york-moors-national-park",
    "<OT237>": "national-employment-savings-trust",
    "<OT248>": "air-accidents-investigation-branch",
    "<OT249>": "rail-accidents-investigation-branch",
    "<OT261>": "office-of-tax-simplification",
    "<OT269>": "equality-and-human-rights-commission",
    "<OT281>": "hm-inspectorate-of-constabulary",
    "<OT284>": "chief-inspector-of-the-uk-border-agency",
    "<OT304>": "the-security-service-mi5",
    "<OT305>": "secret-intelligence-service",
    "<OT306>": "government-communications-headquarters",
    "<OT313>": "british-council",
    "<OT315>": "marshall-aid-commemoration-commission",
    "<OT316>": "westminster-foundation-for-democracy",
    "<OT328>": "official-solicitor-and-public-trustee",
    "<OT342>": "privy-council-office",
    "<OT347>": "hm-crown-prosecution-service-inspectorate",
    "<OT360>": "uk-green-investment-bank",
    "<OT385>": "ofcom",
    "<OT404>": "service-complaints-commissioner",
    "<OT405>": "defence-academy",
    "<OT406>": "service-prosecuting-authority",
    "<OT408>": "defence-press-and-broadcasting-advisory-committee",
    "<OT409>": "royal-navy-submarine-museum",
    "<OT411>": "defence-sixth-form-college",
    "<OT425>": "drinking-water-inspectorate",
    "<OT428>": "boundary-commission-for-northern-ireland",
    "<OT429>": "northern-ireland-human-rights-commission",
    "<OT430>": "parades-commission-for-northern-ireland",
    "<OT431>": "boundary-commission-for-scotland",
    "<OT432>": "boundary-commission-for-wales",
    "<OT433>": "the-adjudicator-s-office",
    "<OT444>": "independent-monitoring-boards-of-prisons-immigration-removal-"
               "centres-and-short-term-holding-rooms",
    "<OT483>": "health-research-authority",
    "<OT484>": "nhs-trust-development-authority",
    "<OT486>": "nhs-blood-and-transplant",
    "<OT488>": "nhs-litigation-authority",
    "<OT492>": "chevening-foundation",
    "<OT494>": "airports-commission",
    "<OT495>": "defence-equipment-and-support",
    "<OT496>": "defence-infrastructure-organisation",
    "<OT498>": "joint-forces-command",
    "<OT502>": "office-for-life-sciences",
    "<OT504>": "chief-fire-and-rescue-adviser-unit",
    "<OT505>": "national-security",
    "<OT506>": "government-equalities-office",
    "<OT507>": "efficiency-and-reform-group",
    "<OT511>": "independent-reviewer-of-terrorism-legislation",
    "<OT512>": "intelligence-services-commissioner",
    "<OT513>": "interception-of-communications-commissioner",
    "<OT514>": "office-for-low-emission-vehicles",
    "<OT515>": "office-of-the-parliamentary-counsel",
    "<OT517>": "commissioner-for-public-appointments",
    "<OT518>": "mckay-commission",
    "<OT519>": "the-parliamentary-and-health-service-ombudsman",
    "<OT520>": "behavioural-insights-team",
    "<OT522>": "advisory-committee-on-clinical-excellence-awards",
    "<OT529>": "nhs-business-services-authority",
    "<OT532>": "prime-ministers-office-10-downing-street",
    "<OT533>": "investigation-into-the-role-of-jimmy-savile-at-broadmoor-"
               "hospital",
    "<OT535>": "border-force",
    "<OT536>": "forensic-science-regulator",
    "<OT537>": "deputy-prime-ministers-office",
    "<OT538>": "health-and-social-care-information-centre",
    "<OT539>": "health-education-england",
    "<OT540>": "infrastructure-uk",
    "<OT554>": "uk-visas-and-immigration",
    "<PB27>": "student-loans-company",
    "<PB28>": "acas",
    "<PB29>": "national-fraud-authority",
    "<PB57>": "marine-management-organisation",
    "<PB118>": "horserace-betting-levy-board",
    "<PB120>": "higher-education-funding-council-for-england",
    "<PB121>": "council-for-science-and-technology",
    "<PB122>": "low-pay-commission",
    "<PB123>": "arts-and-humanities-research-council",
    "<PB124>": "british-hallmarking-council",
    "<PB126>": "construction-industry-training-board",
    "<PB129>": "economic-and-social-research-council",
    "<PB130>": "engineering-and-physical-sciences-research-council",
    "<PB131>": "engineering-construction-industry-training-board",
    "<PB132>": "medical-research-council",
    "<PB133>": "natural-environment-research-council",
    "<PB134>": "office-for-fair-access",
    "<PB135>": "science-and-technology-facilities-council",
    "<PB136>": "technology-strategy-board",
    "<PB137>": "uk-atomic-energy-authority",
    "<PB138>": "uk-commission-for-employment-and-skills",
    "<PB139>": "certification-office",
    "<PB140>": "competition-appeal-tribunal",
    "<PB147>": "capital-for-enterprise-ltd",
    "<PB148>": "central-arbitration-committee",
    "<PB158>": "the-committee-on-standards-in-public-life",
    "<PB160>": "building-regulations-advisory-committee",
    "<PB161>": "homes-and-communities-agency",
    "<PB165>": "arts-council-england",
    "<PB166>": "british-library",
    "<PB167>": "british-museum",
    "<PB168>": "english-heritage",
    "<PB169>": "gambling-commission",
    "<PB170>": "geffrye-museum",
    "<PB171>": "horniman-museum",
    "<PB172>": "imperial-war-museum",
    "<PB174>": "national-gallery",
    "<PB175>": "national-heritage-memorial-fund",
    "<PB176>": "national-lottery-commission",
    "<PB177>": "science-museum-group",
    "<PB178>": "national-museums-liverpool",
    "<PB179>": "national-portrait-gallery",
    "<PB180>": "natural-history-museum",
    "<PB181>": "olympic-delivery-authority",
    "<PB182>": "royal-armouries-museum",
    "<PB183>": "sir-john-soane-s-museum",
    "<PB184>": "sports-grounds-safety-authority",
    "<PB185>": "uk-sport",
    "<PB186>": "visitbritain",
    "<PB187>": "wallace-collection",
    "<PB188>": "big-lottery-fund",
    "<PB189>": "british-film-institute",
    "<PB190>": "sport-england",
    "<PB191>": "committee-on-radioactive-waste-management",
    "<PB192>": "the-fuel-poverty-advisory-group",
    "<PB193>": "nuclear-liabilities-financing-assurance-board",
    "<PB194>": "civil-nuclear-police-authority",
    "<PB195>": "the-coal-authority",
    "<PB196>": "committee-on-climate-change",
    "<PB197>": "nuclear-decommissioning-authority",
    "<PB198>": "consumer-council-for-water",
    "<PB200>": "gangmasters-licensing-authority",
    "<PB201>": "joint-nature-conservation-committee",
    "<PB202>": "natural-england",
    "<PB207>": "health-and-safety-executive",
    "<PB208>": "advisory-committee-on-pesticides",
    "<PB209>": "advisory-committee-on-releases-to-the-environment",
    "<PB210>": "independent-agricultural-appeals-panel",
    "<PB211>": "science-advisory-council",
    "<PB212>": "veterinary-products-committee",
    "<PB213>": "agricultural-land-tribunal",
    "<PB226>": "commonwealth-scholarship-commission-in-the-uk",
    "<PB230>": "equality-2025",
    "<PB231>": "industrial-injuries-advisory-council",
    "<PB232>": "social-security-advisory-committee",
    "<PB233>": "the-pensions-advisory-service",
    "<PB234>": "the-pensions-regulator",
    "<PB235>": "pension-protection-fund-ombudsman",
    "<PB236>": "pensions-ombudsman",
    "<PB246>": "children-and-family-court-advisory-and-support-service",
    "<PB247>": "disabled-persons-transport-advisory-committee",
    "<PB251>": "care-quality-commission",
    "<PB253>": "human-fertilisation-and-embryology-authority",
    "<PB254>": "human-tissue-authority",
    "<PB255>": "monitor",
    "<PB257>": "foreign-compensation-commission",
    "<PB260>": "office-for-budget-responsibility",
    "<PB263>": "office-of-the-immigration-services-commissioner",
    "<PB264>": "security-industry-authority",
    "<PB265>": "serious-organised-crime-agency",
    "<PB266>": "office-of-manpower-economics",
    "<PB270>": "investigatory-powers-tribunal",
    "<PB271>": "advisory-council-on-the-misuse-of-drugs",
    "<PB273>": "police-advisory-board-for-england-and-wales",
    "<PB274>": "technical-advisory-board",
    "<PB275>": "migration-advisory-committee",
    "<PB276>": "national-dna-database-ethics-group",
    "<PB277>": "police-arbitration-tribunal",
    "<PB278>": "police-discipline-appeals-tribunal",
    "<PB290>": "royal-naval-museum",
    "<PB291>": "royal-museums-greenwich",
    "<PB293>": "civil-justice-council",
    "<PB294>": "law-commission",
    "<PB295>": "the-sentencing-council-for-england-and-wales",
    "<PB296>": "victim-s-advisory-panel",
    "<PB297>": "criminal-injuries-compensation-authority",
    "<PB298>": "information-commissioner-s-office",
    "<PB299>": "judicial-appointments-commission",
    "<PB300>": "legal-services-commission",
    "<PB301>": "parole-board",
    "<PB302>": "youth-justice-board-for-england-and-wales",
    "<PB307>": "national-army-museum",
    "<PB308>": "royal-air-force-museum",
    "<PB310>": "royal-marines-museum",
    "<PB311>": "fleet-air-arm-museum",
    "<PB317>": "biotechnology-biological-sciences-research-council",
    "<PB318>": "competition-service",
    "<PB326>": "victims-commissioner",
    "<PB336>": "advisory-committee-on-business-appointments",
    "<PB337>": "boundary-commission-for-england",
    "<PB348>": "pension-protection-fund",
    "<PB349>": "independent-living-fund",
    "<PB350>": "remploy-ltd",
    "<PB353>": "competition-commission",
    "<PB354>": "regulatory-policy-committee",
    "<PB355>": "copyright-tribunal",
    "<PB356>": "consumer-focus",
    "<PB357>": "export-guarantees-advisory-council",
    "<PB358>": "land-registration-rule-committee",
    "<PB366>": "civil-service-commission",
    "<PB367>": "security-vetting-appeals-panel",
    "<PB368>": "review-body-on-senior-salaries",
    "<PB372>": "housing-ombudsman",
    "<PB373>": "leasehold-advisory-service",
    "<PB374>": "london-thames-gateway-development-corporation",
    "<PB375>": "valuation-tribunal-service-for-england-valuation-tribunal-"
               "service",
    "<PB376>": "visitengland",
    "<PB377>": "uk-anti-doping",
    "<PB378>": "victoria-and-albert-museum",
    "<PB380>": "the-theatres-trust",
    "<PB381>": "olympic-lottery-distributor",
    "<PB382>": "the-reviewing-committee-on-the-export-of-works-of-art-and-"
               "objects-of-cultural-interest",
    "<PB383>": "treasure-valuation-committee",
    "<PB392>": "advisory-committee-on-conscientious-objectors",
    "<PB393>": "advisory-group-on-military-medicine",
    "<PB394>": "armed-forces-pay-review-body",
    "<PB395>": "central-advisory-committee-on-pensions-and-compensation",
    "<PB396>": "defence-nuclear-safety-committee",
    "<PB397>": "defence-scientific-advisory-council",
    "<PB399>": "national-employer-advisory-board",
    "<PB400>": "nuclear-research-advisory-council",
    "<PB401>": "oil-and-pipelines-agency",
    "<PB402>": "science-advisory-committee-on-the-medical-implications-of-"
               "less-lethal-weapons",
    "<PB403>": "veterans-advisory-and-pensions-committees-x13",
    "<PB413>": "office-of-the-children-s-commissioner",
    "<PB414>": "school-teachers-review-body",
    "<PB415>": "national-forest-company",
    "<PB420>": "agricultural-wages-committee-x13",
    "<PB421>": "agricultural-dwelling-house-advisory-committees-x16",
    "<PB423>": "plant-varieties-and-seeds-tribunal",
    "<PB426>": "great-britain-china-centre",
    "<PB434>": "royal-mint-advisory-committee",
    "<PB436>": "administrative-justice-and-tribunals-council",
    "<PB437>": "advisory-committees-on-justices-of-the-peace",
    "<PB438>": "the-advisory-council-on-national-records-and-archives",
    "<PB440>": "civil-procedure-rules-committee",
    "<PB441>": "family-justice-council",
    "<PB442>": "family-procedure-rule-committee",
    "<PB443>": "independent-advisory-panel-on-deaths-in-custody",
    "<PB445>": "insolvency-rules-committee",
    "<PB446>": "prison-services-pay-review-body",
    "<PB448>": "tribunal-procedure-committee",
    "<PB456>": "independent-police-complaint-commission",
    "<PB457>": "office-of-surveillance-commissioners",
    "<PB458>": "police-negotiating-board",
    "<PB459>": "directly-operated-railways-limited",
    "<PB460>": "high-speed-two-limited",
    "<PB461>": "british-transport-police-authority",
    "<PB462>": "trinity-house",
    "<PB463>": "northern-lighthouse-board",
    "<PB464>": "passenger-focus",
    "<PB465>": "traffic-commissioners",
    "<PB466>": "railway-heritage-committee",
    "<PB474>": "national-institute-for-clinical-excellence",
    "<PB477>": "review-board-for-government-contracts",
    "<PB481>": "nhs-commissioning-board",
    "<PB491>": "advisory-panel-on-public-sector-information",
    "<PB500>": "the-shareholder-executive",
    "<PB501>": "the-west-northamptonshire-development-corporation",
    "<PB503>": "local-government-ombudsman",
    "<PB508>": "board-of-trustees-of-the-royal-botanic-gardens-kew",
    "<PB509>": "disclosure-and-barring-service",
    "<PB510>": "social-mobility-and-child-poverty-commission",
    "<PB516>": "criminal-procedure-rule-committee",
    "<PB521>": "insolvency-practitioners-tribunal",
    "<PB523>": "administration-of-radioactive-substances-advisory-committee",
    "<PB524>": "british-pharmacopoeia",
    "<PB525>": "commission-on-human-medicines",
    "<PB526>": "committee-on-mutagenicity-of-chemicals-in-food-consumer-"
               "products-and-the-environment",
    "<PB527>": "independent-reconfiguration-panel",
    "<PB530>": "review-body-on-doctors-and-dentists-remuneration",
    "<PB531>": "nhs-pay-review-body",
    "<PB534>": "animals-in-science-committee",
    "<PB542>": "the-office-of-the-schools-commissioner",
    "<PC163>": "architects-registration-board",
    "<PC259>": "uk-financial-investments-limited",
    "<PC343>": "audit-commission",
    "<PC386>": "channel-4",
    "<PC387>": "s4c",
    "<PC388>": "bbc",
    "<PC389>": "historic-royal-palaces",
    "<PC390>": "heritage-lottery-fund",
    "<PC427>": "bbc-world-service",
    "<PC467>": "brb-residuary-ltd",
    "<PC468>": "trust-ports",
    "<PC469>": "civil-aviation-authority",
    "<PC472>": "marine-accident-investigation-branch",
    "<PC493>": "london-and-continental-railways-ltd",
}



from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)

from dshelpers import download_url

from performanceplatform.collector.gcloud.core import (
    nuke_local_database, save_raw_data, aggregate_and_save,
    push_aggregates)

from performanceplatform.collector.gcloud.sales_parser import (
    get_latest_csv_url)

from performanceplatform.collector.write import DataSet


INDEX_URL = ('https://digitalmarketplace.blog.gov.uk'
             '/sales-accreditation-information/')


def main(credentials, data_set_config, query, options, start_at, end_at,
         filename=None):

    nuke_local_database()

    if filename is not None:
        with open(filename, 'r') as f:
            save_raw_data(f)
    else:
        save_raw_data(download_url(get_latest_csv_url(INDEX_URL)))

    aggregate_and_save()

    data_set = DataSet.from_config(data_set_config)
    data_set.empty_data_set()
    push_aggregates(data_set)



from logstash_formatter import LogstashFormatter
import logging
import os
import pdb
import sys
import traceback


def get_log_file_handler(path):
    handler = logging.FileHandler(path)
    handler.setFormatter(logging.Formatter(
        "%(asctime)s [%(levelname)s] -> %(message)s"))
    return handler


def get_json_log_handler(path, app_name, json_fields):
    handler = logging.FileHandler(path)
    formatter = LogstashFormatter()
    formatter.defaults['@tags'] = ['collector', app_name]
    formatter.defaults['@fields'] = json_fields
    handler.setFormatter(formatter)
    return handler


def uncaught_exception_handler(*exc_info):
    text = "".join(traceback.format_exception(*exc_info))
    logging.error("Unhandled exception: %s", text)


def set_up_logging(app_name, log_level, logfile_path, json_fields=None):
    sys.excepthook = uncaught_exception_handler
    logger = logging.getLogger()
    logger.setLevel(log_level)
    logger.addHandler(get_log_file_handler(
        os.path.join(logfile_path, 'collector.log')))
    logger.addHandler(get_json_log_handler(
        os.path.join(logfile_path, 'collector.log.json'),
        app_name,
        json_fields=json_fields if json_fields else {}))
    logger.info("{0} logging started".format(app_name))



from datetime import datetime, timedelta
import json
import logging

from performanceplatform.collector.write import DataSet

from performanceplatform.collector.pingdom.core import Pingdom

_EARLIEST_DATE = datetime(2005, 1, 1)


def main(credentials, data_set_config, query, options, start_at, end_at):
    start_dt, end_dt = parse_time_range(start_at, end_at)

    pingdom = Pingdom(credentials)

    check_name = query['name']
    pingdom_stats = pingdom.stats(check_name, start_dt, end_dt)

    push_stats_to_data_set(
        pingdom_stats,
        check_name,
        data_set_config)


def parse_time_range(start_dt, end_dt):
    """
    Convert the start/end datetimes specified by the user, specifically:
    - truncate any minutes/seconds
    - for a missing end time, use start + 24 hours
    - for a missing start time, use end - 24 hours
    - for missing start and end, use the last 24 hours
    """
    now = datetime.now()

    if start_dt and not end_dt:
        end_dt = now

    elif end_dt and not start_dt:
        start_dt = _EARLIEST_DATE

    elif not start_dt and not end_dt:  # last 24 hours
        end_dt = now
        start_dt = end_dt - timedelta(days=1)

    return tuple(map(truncate_hour_fraction, (start_dt, end_dt)))


def push_stats_to_data_set(pingdom_stats, check_name, data_set_config):
    data_set = DataSet.from_config(data_set_config)
    data_set.post(
        [convert_from_pingdom_to_performanceplatform(thing, check_name) for
         thing in pingdom_stats])


def get_contents_as_json(path_to_file):
    with open(path_to_file) as file_to_load:
        logging.debug(path_to_file)
        return json.load(file_to_load)


def convert_from_pingdom_to_performanceplatform(pingdom_stats, name_of_check):
    timestamp = pingdom_stats['starttime'].isoformat()
    name_for_id = name_of_check.replace(' ', '_')
    return {
        '_id': "%s.%s" % (name_for_id, timestamp),
        '_timestamp': timestamp,
        'avgresponse': pingdom_stats['avgresponse'],
        'uptime': pingdom_stats['uptime'],
        'downtime': pingdom_stats['downtime'],
        'unmonitored': pingdom_stats['unmonitored'],
        'check': name_of_check
    }


def truncate_hour_fraction(a_datetime):
    return a_datetime.replace(minute=0, second=0, microsecond=0)



# Namespace package: https://docs.python.org/2/library/pkgutil.html
from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)

__version__ = "0.1.1"
__author__ = "GDS Developers"
__author_email__ = "performance@digital.cabinet-office.gov.uk"



#!/usr/bin/env python
# encoding: utf-8

from __future__ import unicode_literals

import datetime
import logging
import re

from collections import OrderedDict

import lxml.html
import unicodecsv
from dshelpers import download_url


_ENCODING = 'cp1252'

UNKNOWN_SECTOR_KEY = 'unknown-sector'

_SECTOR_KEY = {
    'Central Government':       'central-gov',

    'Local Government':         'local-gov',

    'Not for Profit':           'not-for-profit',

    'Health':                   'wider-public-sector',
    'Fire and Rescue':          'wider-public-sector',
    'Devolved Administrations': 'wider-public-sector',
    'Education':                'wider-public-sector',
    'Police':                   'wider-public-sector',
    'Defence':                  'wider-public-sector',
    'Wider Public Sector':      'wider-public-sector',
    'Utility (Historic)':       'wider-public-sector',

    'Private Sector':           UNKNOWN_SECTOR_KEY,
}

_SME_LARGE_KEY = {
    'SME': 'sme',
    'Large': 'large',
}

_LOT_KEY = {
    'Infrastructure as a Service (IaaS)': 'iaas',
    'Platform as a Service (PaaS)': 'paas',
    'Software as a Service (SaaS)': 'saas',
    'Specialist Cloud Services': 'css'
}


def get_latest_csv_url(index_page_url):
    """
    Download the index page and extract the URL of the current CSV.
    """
    index_fobj = download_url(index_page_url)
    csv_url = parse_latest_csv_url(index_fobj)
    logging.info("Latest CSV URL: {}".format(csv_url))
    return csv_url


def parse_latest_csv_url(index_fobj):
    root = lxml.html.fromstring(index_fobj.read())
    elements = root.xpath("//*[contains(text(), 'Current')]"
                          "/a[contains(@href, '.csv')]/@href")
    assert len(elements) == 1, elements
    return elements[0]


def process_csv(f):
    """
    Take a file-like object and yield OrderedDicts to be inserted into raw
    spending database.
    """
    reader = unicodecsv.DictReader(f, encoding=_ENCODING)
    for row in reader:
        month, year = parse_month_year(row['Return Month'])

        yield OrderedDict([
            ('customer_name', row['CustomerName']),
            ('supplier_name', row['SupplierName']),
            ('month', month),
            ('year', year),
            ('date', datetime.date(year, month, 1)),
            ('total_ex_vat', parse_price(row['EvidencedSpend'])),
            ('lot', parse_lot_name(row['LotDescription'])),
            ('customer_sector', parse_customer_sector(row['Sector'])),
            ('supplier_type', parse_sme_or_large(row['SME or Large'])),
        ])


def parse_month_year(date_string):
    """
    >>> parse_month_year('01/10/2012')
    (10, 2012)
    """
    match = re.match('\d{2}/(?P<month>\d{2})/(?P<year>\d{4})$',
                     date_string.lower())
    if not match:
        raise ValueError("Not format 'dd/mm/yyyy': '{}'".format(date_string))
    month = int(match.group('month'))
    year = int(match.group('year'))
    return month, year


def parse_price(price_string):
    """
    >>> parse_price('£16,000.00')
    16000.0
    >>> parse_price('-£16,000.00')
    -16000.0
    """
    return float(re.sub('[£,]', '', price_string))

    match = re.match('(?P<amount>-?£[\d,]+(\.\d{2})?)', price_string)
    if not match:
        raise ValueError("Charge not in format '(-)£16,000(.00)' : {}".format(
            repr(price_string)))
    return float(re.sub('[£,]', '', match.group('amount')))


def parse_customer_sector(raw_sector):
    if raw_sector not in _SECTOR_KEY:
        logging.warning('Unknown sector: "{}"'.format(raw_sector))
    return _SECTOR_KEY.get(raw_sector, UNKNOWN_SECTOR_KEY)


def parse_sme_or_large(raw_sme_large):
    try:
        return _SME_LARGE_KEY[raw_sme_large]
    except KeyError:
        raise RuntimeError('Unknown sme/large: "{}"'.format(raw_sme_large))


def parse_lot_name(raw_lot_name):
    return _LOT_KEY[raw_lot_name]



import io
import os
import re
from setuptools import setup, find_packages

from performanceplatform import collector

# multiprocessing and logging don't get on with each other in Python
# vesions < 2.7.4. The following unused import is a workaround. See:
# http://bugs.python.org/issue15881#msg170215
import multiprocessing


def _read(fname, fail_silently=False):
    """
    Read the content of the given file. The path is evaluated from the
    directory containing this file.
    """
    try:
        filepath = os.path.join(os.path.dirname(__file__), fname)
        with io.open(filepath, 'rt', encoding='utf8') as f:
            return f.read()
    except:
        if not fail_silently:
            raise
        return ''


def _get_version():
    data = _read(
        'performanceplatform/collector/__init__.py'
    )
    version = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        data,
        re.M | re.I
    ).group(1).strip()
    return version.encode('ascii')


def _get_requirements(fname):
    """
    Create a list of requirements from the output of the pip freeze command
    saved in a text file.
    """
    packages = _read(fname).split('\n')
    packages = (p.strip() for p in packages)
    packages = (p for p in packages if p and not p.startswith('#'))
    return list(packages)


def _get_long_description():
    return _read('README.rst')


if __name__ == '__main__':
    setup(
        name='performanceplatform-collector',
        version=_get_version(),
        packages=find_packages(exclude=['test*']),

        # metadata for upload to PyPI
        author=collector.__author__,
        author_email=collector.__author_email__,
        maintainer='Government Digital Service',
        url='https://github.com/alphagov/performanceplatform-collector',

        description='performanceplatform-collector: tools for sending'
            'data to the Performance Platform',
        long_description=_get_long_description(),
        license='MIT',
        keywords='api data performance_platform',

        install_requires=_get_requirements('requirements.txt'),
        tests_require=_get_requirements('requirements_for_tests.txt'),

        test_suite='nose.collector',

        entry_points={
            'console_scripts': [
                'pp-collector=performanceplatform.collector.main:main'
            ]
        }
    )



class Comment(object):

    """
    Plugin which ignores all of its parameters. Useful for putting comments in
    the list of plugins.
    """

    def __init__(self, *args):
        pass

    def __call__(self, documents):
        return documents



import base64
from datetime import date, timedelta
import gapy.client

from performanceplatform.collector.write import DataSet
from performanceplatform.utils.http_with_backoff import HttpWithBackoff

ga_date_keys = ['day', 'month', 'year']


def parse_query(query):
    if 'metric' not in query or not query['metric']:
        raise Exception('Metric required')
    else:
        if 'dimensions' in query:
            query['dimensions'].extend(ga_date_keys)
        else:
            query['dimensions'] = ga_date_keys

    return query


def get_date():
    inclusive_end = date.today()
    exclusive_end = inclusive_end - timedelta(days=1)
    middle = inclusive_end - timedelta(weeks=1)
    start = inclusive_end - timedelta(weeks=2)
    return (start, middle, exclusive_end)


def assign_day_to_week(day, month, year, dates):

    start, middle, end = dates

    d = date(int(year), int(month), int(day))

    return 2 if (d >= middle) else 1


def get_trends(data):

    for key in data:
        numerator = data[key]['week2'] - data[key]['week1']
        denominator = data[key]['week1']

        percent_change = ((float(numerator) / float(denominator)) * 100)
        data[key]['percent_change'] = percent_change

    return data


def sum_data(data, metric, collapse_key, dates, floor):

    collapsed = {}

    for row in data:

        dimensions = row['dimensions']
        k = dimensions[collapse_key]

        if k not in collapsed:
            d = {}
            for dim in dimensions:
                if dim not in ga_date_keys:
                    d[dim] = dimensions[dim]
            d['week1'] = 0
            d['week2'] = 0
            collapsed[k] = d

        date_list = [dimensions['day'],
                     dimensions['month'], dimensions['year']]
        if '(other)' in date_list:
            continue
        week = 'week%d' % assign_day_to_week(dimensions['day'],
                                             dimensions['month'],
                                             dimensions['year'], dates)

        # Use the shortest common path.
        if dimensions['pagePath'] in collapsed[k]['pagePath']:
            collapsed[k]['pagePath'] = dimensions['pagePath']

        collapsed[k][week] += int(row['metrics'][metric])

    for key in collapsed:
        for week in ['week1', 'week2']:
            if collapsed[key][week] < floor:
                collapsed[key][week] = floor

    return collapsed


def encode_id(id):
    id_bytes = id.encode('utf-8')
    return base64.urlsafe_b64encode(id_bytes)


def flatten_data_and_assign_ids(data):

    flattened = []

    for key in data:
        data[key]['_id'] = encode_id(data[key]['pagePath'])
        flattened.append(data[key])

    return flattened


def main(credentials, data_set_config, query, options, start_at, end_at):

    credentials = credentials
    client = gapy.client.from_secrets_file(
        credentials['CLIENT_SECRETS'],
        storage_path=credentials['STORAGE_PATH'],
        http_client=HttpWithBackoff(),
    )

    ga_query = parse_query(query)

    collapse_key = "pageTitle"

    (start, middle, end) = get_date()

    data = client.query.get(
        ga_query['id'],
        start,
        end,
        [ga_query['metric']],
        ga_query['dimensions'],
        ga_query['filters'] if 'filters' in ga_query else None
    )

    collapsed_data = sum_data(data, ga_query['metric'], collapse_key,
                              (start, middle, end), 500)
    trended_data = get_trends(collapsed_data)
    flattened_data = flatten_data_and_assign_ids(trended_data)

    data_set = DataSet.from_config(data_set_config)

    data_set.post(flattened_data)



from __future__ import division

from itertools import groupby


class AggregateKey(object):

    """
    Given a set of documents, find all records with equal "key" where the "key"
    is all values which are not being aggregated over.

    For example:

        doc1 = {"foo": "bar", "count": 2}
        doc2 = {"foo": "bar", "count": 2}

        plugin = AggregateKey(aggregate_count("count"))

        plugin([doc1, doc2]) => {"foo": "bar", "count": 4}

    Aggregate rates are also supported through
    `aggregate_rate(rate_key, count_key)`.
    """

    def __init__(self, *aggregations):
        self.aggregations = aggregations

    def __call__(self, documents):
        first = documents[0]

        aggregate_keys = [k for k, _ in self.aggregations]
        groupkeys = set(first) - set(aggregate_keys)

        def key(doc):
            return tuple(doc[key] for key in groupkeys)

        return [make_aggregate(grouped, self.aggregations)
                for grouped in group(documents, key)]


def group(iterable, key):
    """
    groupby which sorts the input, discards the key and returns the output
    as a sequence of lists.
    """
    for _, grouped in groupby(sorted(iterable, key=key), key=key):
        yield list(grouped)


def aggregate_count(keyname):
    """
    Straightforward sum of the given keyname.
    """
    def inner(docs):
        return sum(doc[keyname] for doc in docs)

    return keyname, inner


def aggregate_rate(rate_key, count_key):
    """
    Compute an aggregate rate for `rate_key` weighted according to
    `count_rate`.
    """
    def inner(docs):
        total = sum(doc[count_key] for doc in docs)
        weighted_total = sum(doc[rate_key] * doc[count_key] for doc in docs)
        total_rate = weighted_total / total
        return total_rate

    return rate_key, inner


def make_aggregate(docs, aggregations):
    """
    Given `docs` and `aggregations` return a single document with the
    aggregations applied.
    """
    new_doc = dict(docs[0])

    for keyname, aggregation_function in aggregations:
        new_doc[keyname] = aggregation_function(docs)

    return new_doc


def test_make_aggregate_sum():
    """
    test_make_aggregate_sum()

    Straight test that summation works over the field specified in
    aggregate_count(keyname).
    """
    from nose.tools import assert_equal

    doc1 = {"a": 2, "b": 2, "c": 2, "visits": 201}
    doc2 = {"a": 2, "b": 2, "c": 2, "visits": 103}
    docs = [doc1, doc2]

    aggregate_doc = make_aggregate(docs, [aggregate_count("visits")])
    expected_aggregate = {"a": 2, "b": 2, "c": 2, "visits": 304}
    assert_equal(aggregate_doc, expected_aggregate)


def test_make_aggregate_rate():
    """
    test_make_aggregate_rate()

    Test that aggregation works when there are two fields summed over, and
    that rate aggregation works correctly. It isn't easy to test independently
    because one must aggregate over the two fields to get the correct result.
    """
    from nose.tools import assert_equal

    doc1 = {"a": 2, "b": 2, "c": 2, "visits": 100, "rate": 0.25}
    doc2 = {"a": 2, "b": 2, "c": 2, "visits": 100, "rate": 0.75}
    docs = [doc1, doc2]

    aggregate_doc = make_aggregate(docs, [aggregate_count("visits"),
                                          aggregate_rate("rate", "visits")])
    expected_aggregate = {
        "a": 2, "b": 2, "c": 2,
        "visits": 200,
        "rate": (0.25 * 100 + 0.75 * 100) / (100 + 100)}

    assert_equal(aggregate_doc, expected_aggregate)


def test_AggregateKeyPlugin():
    """
    test_AggregateKeyPlugin()

    Test that the AggregateKey class behaves as a plugin correctly.
    """
    from nose.tools import assert_equal

    doc1 = {"a": 2, "b": 2, "c": 2, "visits": 100, "rate": 0.25}
    doc2 = {"a": 2, "b": 2, "c": 2, "visits": 100, "rate": 0.75}
    docs = [doc1, doc2]

    plugin = AggregateKey(aggregate_count("visits"),
                          aggregate_rate("rate", "visits"))

    output_docs = plugin(docs)

    expected_aggregate = {
        "a": 2, "b": 2, "c": 2,
        "visits": 200,
        "rate": (0.25 * 100 + 0.75 * 100) / (100 + 100)}

    assert_equal(output_docs, [expected_aggregate])



from datetime import datetime, time, timedelta, date
import pytz


MONDAY = 0


def to_datetime(a_date):
    return datetime.combine(a_date, time(0)).replace(tzinfo=pytz.UTC)


def to_date(a_datetime):
    if a_datetime is None:
        return None
    elif isinstance(a_datetime, datetime):
        return a_datetime.date()
    elif isinstance(a_datetime, date):
        return a_datetime
    else:
        raise ValueError("{0!r} ({1}) isn't a date or datetime"
                         .format(a_datetime, type(a_datetime)))


def to_utc(a_datetime):
    return a_datetime.astimezone(pytz.UTC)


def period_range(start_date, end_date):
    start_date = to_date(start_date) or a_week_ago()
    end_date = to_date(end_date) or a_week_ago()

    if start_date > end_date:
        raise ValueError("Bad period: !(start_date={0} <= end_date={1})"
                         .format(start_date, end_date))

    if start_date.weekday != MONDAY:
        start_date = start_date - timedelta(days=start_date.weekday())

    period = timedelta(days=7)
    while start_date <= end_date:
        yield (start_date, start_date + timedelta(days=6))
        start_date += period


def a_week_ago():
    return date.today() - timedelta(days=7)



import base64
import datetime

import pytz


def to_utc(a_datetime):
    return a_datetime.astimezone(pytz.UTC)


def value_id(value):
    value_bytes = value.encode('utf-8')
    return base64.urlsafe_b64encode(value_bytes), value_bytes


class ComputeIdFrom(object):

    def __init__(self, *fields):
        self.fields = fields

    def __call__(self, documents):
        for document in documents:

            id_parts = (stringify(document[field]) for field in self.fields)
            id_string = "_".join(id_parts)

            _id, humanId = value_id(id_string)

            document['_id'] = _id
            document['humanId'] = humanId

        return documents


def stringify(item):
    if isinstance(item, datetime.datetime):
        return _format(item)
    return u"{0!s}".format(item)


def _format(timestamp):
    return to_utc(timestamp).strftime("%Y%m%d%H%M%S")
